from datetime import date
from typing import Any

from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.utils.functional import cached_property

from .checks import RiskCheck, evaluate_as_task
from .rules import RiskRule, RuleCheckedObjectRelationship


class RiskCheckMixin(models.Model):
    """
    A utility mixin to inherit from when a model proposes a risk check workflow on one of its field
    """

    id: int

    def _get_checked_object_field_name(self) -> str:
        """
        Return the object representing the checked object
        Returns NotImplementedError by default
        """
        raise NotImplementedError()

    @cached_property
    def content_type(self) -> ContentType:
        return ContentType.objects.get_for_model(self)

    @cached_property
    def checked_object(self) -> Any:
        return getattr(self, self._get_checked_object_field_name())

    @property
    def checks(self) -> models.QuerySet[RiskCheck]:
        """
        Returned the check triggered by this object (self)
        Returns: A queryset of RiskCheck
        """
        return RiskCheck.objects.filter(trigger_content_type=self.content_type, trigger_id=self.id)

    @cached_property
    def has_assigned_active_rules(self) -> bool:
        """
        Return True if an enabled and active rule is available for the assigned checked object
        """
        relationships = RuleCheckedObjectRelationship.objects.get_for_object(
            self.checked_object, rule__is_enable=True, rule__only_passive_check_allowed=False
        )
        return RiskRule.objects.filter(id__in=relationships.distinct("rule").values("rule")).exists()

    @property
    def has_all_check_completed_and_succeed(self) -> bool:
        """
        Return True if checks are available and they all succeed
        """
        return (
            self.checks.exists()
            and not self.checks.exclude(
                status__in=[RiskCheck.CheckStatus.SUCCESS, RiskCheck.CheckStatus.WARNING]
            ).exists()
        )

    @property
    def has_all_check_completed(self) -> bool:
        """
        Return True if checks are available and they all succeed
        """
        return (
            self.checks.exists()
            and not self.checks.filter(
                status__in=[RiskCheck.CheckStatus.RUNNING, RiskCheck.CheckStatus.PENDING]
            ).exists()
        ) or not self.checks.exists()

    @property
    def has_no_rule_or_all_checked_succeed(self) -> bool:
        return (
            self.has_assigned_active_rules and self.has_all_check_completed_and_succeed
        ) or not self.has_assigned_active_rules

    def get_worst_check_status(self) -> RiskCheck.CheckStatus:
        status_ordered = [
            RiskCheck.CheckStatus.FAILED,
            RiskCheck.CheckStatus.WARNING,
            RiskCheck.CheckStatus.RUNNING,
            RiskCheck.CheckStatus.PENDING,
        ]
        for status in status_ordered:
            if self.checks.filter(status=status).exists():
                return status
        return RiskCheck.CheckStatus.SUCCESS

    def evaluate_active_rules(self, evaluation_date: date, *dto, asynchronously: bool = True):
        for rel in RuleCheckedObjectRelationship.objects.get_for_object(
            self.checked_object, rule__is_enable=True, rule__only_passive_check_allowed=False
        ):
            check = RiskCheck.objects.update_or_create(
                rule_checked_object_relationship=rel,
                evaluation_date=evaluation_date,
                trigger_content_type=self.content_type,
                trigger_id=self.id,
                defaults={"status": RiskCheck.CheckStatus.PENDING},
            )[0]
            if asynchronously:
                transaction.on_commit(
                    lambda: evaluate_as_task.delay(
                        check.id, *dto, override_incident=True, ignore_informational_threshold=True
                    )
                )
            else:
                check.evaluate(*dto, override_incident=True, ignore_informational_threshold=True)

    class Meta:
        abstract = True



from django.utils.translation import gettext as _
from wbcore.metadata.configs.titles import TitleViewConfig


class ComplianceTypeTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Compliance Administrator: {{name}}")

    def get_list_title(self):
        return _("Compliance Administrators")

    def get_create_title(self):
        return _("New Compliance Administrator")



from wbcore import filters as wb_filters

from wbcompliance.models.risk_management import RiskCheck


class RiskCheckFilterSet(wb_filters.FilterSet):
    triggers = wb_filters.MultipleChoiceContentTypeFilter(
        label="Triggerers",
        field_name="triggers",
        object_id_label="trigger_id",
        content_type_label="trigger_content_type",
        distinct=True,
    )

    class Meta:
        model = RiskCheck

        fields = {
            "rule_checked_object_relationship": ["exact"],
            "creation_datetime": ["gte", "exact", "lte"],
            "evaluation_date": ["exact"],
            "status": ["exact"],
        }



import factory

from wbcompliance.models.risk_management.checks import RiskCheck


class RiskCheckFactory(factory.django.DjangoModelFactory):
    rule_checked_object_relationship = factory.SubFactory(
        "wbcompliance.factories.risk_management.rules.RuleCheckedObjectRelationshipFactory"
    )
    evaluation_date = factory.Faker("date_object")

    class Meta:
        model = RiskCheck



from contextlib import suppress

from django.contrib.contenttypes.models import ContentType
from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.signals import add_additional_resource, add_instance_button

from wbcompliance.models.risk_management.checks import RiskCheck


class RiskCheckViewSetMixin:
    @classmethod
    def _get_risk_checks_button_title(cls) -> str:
        return "Checks"

    @classmethod
    def _add_risk_check_button(cls, sender, many, *args, view=None, **kwargs):
        icon = WBIcon.CONFIRM.icon
        with suppress(AssertionError):
            if view and (obj := view.get_object()):
                icon = RiskCheck.CheckStatus[obj.get_worst_check_status()].icon
        return bt.WidgetButton(key="checks", label=cls._get_risk_checks_button_title(), icon=icon)

    @classmethod
    def _add_risk_check_add_additional_resource(
        cls, sender, serializer, instance, request, user, view=None, is_list=False, **kwargs
    ):
        if not is_list and instance.checks.exists() and (content_type := ContentType.objects.get_for_model(instance)):
            return {
                "checks": f'{reverse("wbcompliance:riskcheck-list", args=[], request=request)}?triggers=[[{content_type.id},{instance.id}]]'
            }
        return {}

    def options(self, request, *args, **kwargs):
        """
        Handler method for HTTP 'OPTIONS' request.
        """
        add_instance_button.connect(
            self._add_risk_check_button,
            sender=self.__class__,
            dispatch_uid="wbcompliance_add_instance_button_riskcheck",
        )
        return super().options(request, *args, **kwargs)

    def get_serializer_class(self):
        serializer_class = super().get_serializer_class()
        add_additional_resource.connect(
            self._add_risk_check_add_additional_resource,
            sender=serializer_class,
            dispatch_uid="wbcompliance_add_additional_resource_riskcheck",
        )
        return serializer_class
